# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=unsubscriptable-object

"""
Shared utilities.
"""

import asyncio
import logging
import re
import sys
import time
from asyncio import TimeoutError  # pylint: disable=redefined-builtin
from logging import FileHandler, StreamHandler
from pathlib import Path
from typing import Optional, Union, cast

from websockets.legacy.protocol import WebSocketCommonProtocol
from websockets.typing import Data

from .interface import Message, MessageType, ProcessingError, RequestFailedContext

log = logging.getLogger("apologies.util")


def homedir() -> str:
    """Get the current user's home directory."""
    return str(Path.home())


def mask(data: Optional[Union[str, bytes]]) -> str:
    """Mask the player id in JSON data, since it's a secret we don't want logged."""
    decoded = "" if not data else data.decode("utf-8") if isinstance(data, bytes) else data
    return re.sub(r'"player_id" *: *"[^"]+"', r'"player_id": "<masked>"', decoded)


def extract(data: Union[str, Message, Data]) -> Message:
    message = Message.for_json(str(data))
    if message.message == MessageType.REQUEST_FAILED:
        context = cast(RequestFailedContext, message.context)
        raise ProcessingError(reason=context.reason, comment=context.comment, handle=context.handle)
    return message


async def close(websocket: WebSocketCommonProtocol) -> None:
    """Close a websocket."""
    log.debug("Closing websocket: %s", id(websocket))
    await websocket.close()


async def send(websocket: WebSocketCommonProtocol, message: Union[str, Message]) -> None:
    """Send a response to a websocket."""
    if message:
        data = message.to_json() if isinstance(message, Message) else message
        log.debug("Sending message to websocket: %s\n%s", id(websocket), mask(data))
        await websocket.send(data)


async def receive(websocket: WebSocketCommonProtocol, timeout_sec: Optional[int] = None) -> Optional[Message]:
    try:
        data = await websocket.recv() if not timeout_sec else await asyncio.wait_for(websocket.recv(), timeout=timeout_sec)
        log.debug("Received raw data for websocket %s:\n%s", id(websocket), mask(data))
        return extract(data)
    except TimeoutError:
        log.debug("Timed out waiting for raw data for websocket %s", id(websocket))
        return None


def setup_logging(quiet: bool, verbose: bool, debug: bool, logfile_path: Optional[str] = None) -> None:
    """Set up Python logging."""
    logger = logging.getLogger("apologies")
    logger.setLevel(logging.DEBUG)
    handler: StreamHandler = FileHandler(logfile_path) if logfile_path else StreamHandler(sys.stdout)  # type: ignore
    formatter = logging.Formatter(fmt="%(asctime)sZ --> [%(levelname)-7s] %(message)s")
    formatter.converter = time.gmtime  # type: ignore
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    if quiet:
        handler.setLevel(logging.ERROR)
    if verbose or debug:
        handler.setLevel(logging.DEBUG)
    if debug:
        wslogger = logging.getLogger("websockets")
        wslogger.setLevel(logging.INFO)
        wslogger.addHandler(handler)
    logger.addHandler(handler)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name,wildcard-import,protected-access,too-many-lines

import random
from typing import Dict, List
from unittest.mock import MagicMock, patch

import pendulum
import pytest
from apologies.game import GameMode, PlayerColor
from apologies.rules import Rules
from apologies.source import NoOpInputSource, RewardV1InputSource
from ordered_set import OrderedSet

from apologiesserver.interface import *
from apologiesserver.manager import (
    _MANAGER,
    _NAMES,
    CurrentTurn,
    StateManager,
    TrackedEngine,
    TrackedGame,
    TrackedPlayer,
    TrackedWebsocket,
    manager,
)

from .util import random_string, to_date


def create_test_player(player_id="id", handle="handle"):
    return TrackedPlayer(player_id=player_id, websocket=MagicMock(), handle=handle)


def create_test_game():
    engine = MagicMock()
    return TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PRIVATE, [], engine=engine)


def check_fully_joined(players: int, game_players: int) -> bool:
    game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PRIVATE, [])
    game.players = players
    game.game_players = {random_string(): MagicMock() for _ in range(game_players)}
    return game.is_fully_joined()


def check_is_in_progress(advertised: bool, playing: bool) -> bool:
    game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PRIVATE, ["bender", "fry"])
    game.is_advertised = MagicMock()  # type: ignore
    game.is_playing = MagicMock()  # type: ignore
    game.is_advertised.return_value = advertised
    game.is_playing.return_value = playing
    return game.is_in_progress()


def check_is_viable(advertised: bool, available: int) -> bool:
    game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PRIVATE, [])
    game.is_advertised = MagicMock()  # type: ignore
    game.get_available_player_count = MagicMock()  # type: ignore
    game.is_advertised.return_value = advertised
    game.get_available_player_count.return_value = available
    return game.is_viable()


class TestFunctions:
    """
    Test Python functions.
    """

    def test_manager(self):
        assert _MANAGER is not None
        assert manager() is _MANAGER


class TestTrackedWebsocket:
    """
    Test the TrackedWebsocket class.
    """

    # noinspection PyTypeChecker
    def test_mark_active(self):
        now = pendulum.now()
        tracked = TrackedWebsocket(MagicMock())
        tracked.last_active_date = None
        tracked.activity_state = None
        tracked.mark_active()
        assert tracked.last_active_date >= now
        assert tracked.activity_state == ActivityState.ACTIVE

    def test_mark_idle(self):
        tracked = TrackedWebsocket(MagicMock())
        tracked.activity_state = None
        tracked.mark_idle()
        assert tracked.activity_state == ActivityState.IDLE

    def test_mark_inactive(self):
        tracked = TrackedWebsocket(MagicMock())
        tracked.activity_state = None
        tracked.mark_inactive()
        assert tracked.activity_state == ActivityState.INACTIVE


class TestTrackedPlayer:
    """
    Test the TrackedPlayer class.
    """

    def test_for_context(self):
        now = pendulum.now()
        websocket = MagicMock()
        player = TrackedPlayer.for_context("id", websocket, "handle")
        assert player.player_id == "id"
        assert player.handle == "handle"
        assert player.websocket is websocket
        assert player.registration_date >= now
        assert player.last_active_date >= now
        assert player.activity_state == ActivityState.ACTIVE
        assert player.connection_state == ConnectionState.CONNECTED
        assert player.player_state == PlayerState.WAITING
        assert player.game_id is None

    def test_to_registered_player(self):
        websocket = MagicMock()
        player = TrackedPlayer(player_id="id", websocket=websocket, handle="handle", game_id="game_id")
        registered = player.to_registered_player()
        assert registered.handle == "handle"
        assert registered.registration_date == player.registration_date
        assert registered.last_active_date == player.last_active_date
        assert registered.connection_state == player.connection_state
        assert registered.activity_state == player.activity_state
        assert registered.player_state == player.player_state
        assert registered.game_id == player.game_id

    # noinspection PyTypeChecker
    def test_mark_active(self):
        now = pendulum.now()
        player = create_test_player()
        player.last_active_date = None
        player.activity_state = None
        player.connection_state = None
        player.mark_active()
        assert player.last_active_date >= now
        assert player.activity_state == ActivityState.ACTIVE
        assert player.connection_state == ConnectionState.CONNECTED

    def test_mark_idle(self):
        player = create_test_player()
        player.activity_state = None
        player.mark_idle()
        assert player.activity_state == ActivityState.IDLE

    def test_mark_inactive(self):
        player = create_test_player()
        player.activity_state = None
        player.mark_inactive()
        assert player.activity_state == ActivityState.INACTIVE

    def test_mark_joined(self):
        game = MagicMock(game_id="game_id")
        player = create_test_player()
        player.game_id = None
        player.player_state = None
        player.mark_joined(game)
        assert player.game_id == "game_id"
        assert player.player_state == PlayerState.JOINED

    def test_mark_playing(self):
        player = create_test_player()
        player.player_state = None
        player.mark_playing()
        assert player.player_state == PlayerState.PLAYING

    def test_mark_quit(self):
        player = create_test_player()
        player.game_id = "game_id"
        player.player_state = None
        player.mark_quit()
        assert player.game_id is None
        assert player.player_state == PlayerState.WAITING  # they move through QUIT and back to WAITING

    def test_mark_disconnected(self):
        player = create_test_player()
        player.game_id = "game_id"
        player.activity_state = None
        player.connection_state = None
        player.player_state = None
        player.mark_disconnected()
        assert player.game_id is None
        assert player.websocket is None
        assert player.activity_state == ActivityState.IDLE
        assert player.connection_state == ConnectionState.DISCONNECTED
        assert player.player_state == PlayerState.WAITING  # they move through QUIT and back to WAITING


class TestCurrentTurn:
    """
    Test the CurrentTurn class.
    """

    def test_draw_again(self):
        current = CurrentTurn(handle="handle", color=PlayerColor.RED, view=None, movelist=None, movedict=None)
        move = MagicMock(id="move_id")
        view = MagicMock()
        engine = MagicMock()
        engine.game.create_player_view.return_value = view
        engine.construct_legal_moves.return_value = (None, [move])
        turn = current.draw_again(engine)
        engine.game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine.construct_legal_moves.assert_called_once_with(view)
        assert turn.handle == "handle"
        assert turn.color == PlayerColor.RED
        assert turn.view is view
        assert turn.movelist == [move]
        assert turn.movedict == {"move_id": move}

    def test_next_player(self):
        move = MagicMock(id="move_id")
        character = MagicMock()
        character.configure_mock(name="handle")  # the "name" attribute means something special in the constructor
        view = MagicMock()
        engine = MagicMock()
        engine.next_turn.return_value = (PlayerColor.RED, character)
        engine.game.create_player_view.return_value = view
        engine.construct_legal_moves.return_value = (None, [move])
        turn = CurrentTurn.next_player(engine)
        engine.next_turn.assert_called_once()
        engine.game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine.construct_legal_moves.assert_called_once_with(view)
        assert turn.handle == "handle"
        assert turn.color == PlayerColor.RED
        assert turn.view is view
        assert turn.movelist == [move]
        assert turn.movedict == {"move_id": move}


class TestTrackedEngine:
    """
    Test the TrackedEngine class.
    """

    # noinspection PyTypeChecker
    # noinspection PyUnresolvedReferences
    @patch("apologiesserver.manager.CurrentTurn")
    def test_start_game(self, current_turn):
        replacement = MagicMock()
        current_turn.next_player.return_value = replacement
        mode = GameMode.STANDARD
        handles = ["leela", "bender"]
        engine = TrackedEngine()
        engine._engine = MagicMock()
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.start_game(mode, handles=handles)  # there's already an engine
        engine._engine = None
        engine._colors = {}
        engine._current = None
        colors = engine.start_game(mode, handles)
        assert colors == engine._colors and colors is not engine._colors  # it's a copy
        assert engine._engine.mode == mode
        assert len(engine._engine.characters) == 2
        assert engine._engine.characters[0].name == "leela"
        assert isinstance(engine._engine.characters[0].source, NoOpInputSource)
        assert engine._engine.characters[1].name == "bender"
        assert isinstance(engine._engine.characters[1].source, NoOpInputSource)
        assert engine._engine.started is True
        assert "leela" in engine._colors and "bender" in engine._colors and engine._colors["leela"] != engine._colors["bender"]
        assert engine._current is replacement
        current_turn.next_player.assert_called_with(engine._engine)

    def test_stop_game(self):
        engine = TrackedEngine()
        engine._engine = MagicMock()
        engine._colors = {"a": "b"}
        engine._current = MagicMock()
        engine.stop_game()
        assert engine._engine is None
        assert engine._colors == {}
        assert engine._current is None

    def test_get_next_turn(self):
        engine = TrackedEngine()
        engine._current = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_next_turn()
        engine._current = MagicMock(handle="handle")
        assert engine.get_next_turn() == "handle"

    def test_get_legal_moves(self):
        move = MagicMock()
        engine = TrackedEngine()
        engine._current = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_legal_moves("handle")  # no current turn
        engine._current = MagicMock(handle="handle")
        engine._current.movelist = [move]
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_legal_moves("bogus")  # unknown handle
        assert engine.get_legal_moves("handle") == [move]

    def test_get_player_view(self):
        view = MagicMock()
        engine = TrackedEngine()
        engine._engine = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_player_view("handle")  # no engine
        engine._engine = MagicMock()
        engine._engine.game.create_player_view.return_value = view
        engine._colors = {"handle": PlayerColor.RED}
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_player_view("bogus")  # unknown handle
        assert engine.get_player_view("handle") is view
        engine._engine.game.create_player_view.assert_called_once_with(PlayerColor.RED)

    def test_get_recent_history(self):
        engine = TrackedEngine()
        engine._engine = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_recent_history(10)  # no engine
        engine._engine = MagicMock()
        engine._engine.game = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.get_recent_history(10)  # no game
        engine._engine.game = MagicMock(history=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        assert engine.get_recent_history(10) == [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

    def test_is_move_pending(self):
        engine = TrackedEngine()
        engine._current = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.is_move_pending("handle")  # no current turn
        engine._current = MagicMock(handle="handle")
        assert engine.is_move_pending("bogus") is False
        assert engine.is_move_pending("handle") is True

    def test_is_legal_move(self):
        move = MagicMock()
        engine = TrackedEngine()
        engine._current = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.is_legal_move("handle", "move_id")  # no current turn
        engine._current = MagicMock(handle="handle")
        engine._current.movedict = {"move_id": move}
        assert engine.is_legal_move("bogus", "move_id") is False  # handle doesn't match current turn
        assert engine.is_legal_move("handle", "bogus") is False  # move isn't in the current turn's list
        assert engine.is_legal_move("handle", "move_id") is True

    def test_execute_move_validations(self):
        move = MagicMock()
        engine = TrackedEngine()
        engine._engine = None
        engine._current = None
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.execute_move("handle", "move_id")  # no engine
        engine._engine = MagicMock()
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.execute_move("handle", "move_id")  # no current turn
        engine._current = MagicMock(handle="handle")
        engine._current.movedict = {"move_id": move}
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.execute_move("bogus", "move_id")  # handle doesn't match current turn
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.execute_move("handle", "move_id")  # can't find color mapping for handle
        engine._colors = {"handle": PlayerColor.RED}
        with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
            engine.execute_move("handle", "bogus")  # unknown move

    def test_execute_move_completed(self):
        character = MagicMock()
        character.configure_mock(name="handle")  # the "name" attribute means something special in the constructor
        player = MagicMock(color=PlayerColor.YELLOW, turns=42)
        move = MagicMock()
        current = MagicMock(handle="handle")
        engine = TrackedEngine()
        engine._engine = MagicMock(completed=True)
        engine._current = current
        engine._current.movedict = {"move_id": move}
        engine._colors = {"handle": PlayerColor.RED}
        engine._engine.winner.return_value = (character, player)  # only set if the game is completed
        engine._engine.execute_move.return_value = True  # turn is always true if the game is completed
        assert engine.execute_move("handle", "move_id") == (True, "handle", "Player handle won after 42 turns")
        assert engine._current is None
        engine._engine.execute_move.assert_called_once_with(PlayerColor.RED, move)

    def test_execute_move_not_done(self):
        character = MagicMock()
        character.configure_mock(name="handle")  # the "name" attribute means something special in the constructor
        move = MagicMock()
        current = MagicMock(handle="handle")
        again = MagicMock()
        engine = TrackedEngine()
        engine._engine = MagicMock(completed=False)
        engine._current = current
        engine._current.draw_again.return_value = again
        engine._current.movedict = {"move_id": move}
        engine._colors = {"handle": PlayerColor.RED}
        engine._engine.winner.return_value = None  # only set if the game is completed
        engine._engine.execute_move.return_value = False  # player's turn is not done yet, so current does not change
        assert engine.execute_move("handle", "move_id") == (False, None, None)
        assert engine._current is again
        current.draw_again.assert_called_once_with(engine._engine)
        engine._engine.execute_move.assert_called_once_with(PlayerColor.RED, move)

    @patch("apologiesserver.manager.CurrentTurn")
    def test_execute_move_done(self, current_turn):
        character = MagicMock()
        character.configure_mock(name="handle")  # the "name" attribute means something special in the constructor
        move = MagicMock()
        current = MagicMock(handle="handle")
        replacement = MagicMock()
        current_turn.next_player.return_value = replacement
        engine = TrackedEngine()
        engine._engine = MagicMock(completed=False)
        engine._current = current
        engine._current.movedict = {"move_id": move}
        engine._colors = {"handle": PlayerColor.RED}
        engine._engine.winner.return_value = None  # only set if the game is completed
        engine._engine.execute_move.return_value = True  # player's turn is now done, so current will change
        assert engine.execute_move("handle", "move_id") == (False, None, None)
        assert engine._current is replacement
        engine._engine.execute_move.assert_called_once_with(PlayerColor.RED, move)


# pylint: disable=too-many-public-methods
class TestTrackedGame:
    """
    Test the TrackedGame class.
    """

    def test_for_context(self):
        now = pendulum.now()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["a", "b"])
        game = TrackedGame.for_context("handle", "game_id", context)
        assert game.game_id == "game_id"
        assert game.advertiser_handle == "handle"
        assert game.name == "name"
        assert game.mode == GameMode.STANDARD
        assert game.players == 3
        assert game.visibility == Visibility.PUBLIC
        assert game.invited_handles == ["a", "b"]
        assert game.advertised_date >= now
        assert game.last_active_date >= now
        assert game.started_date is None
        assert game.completed_date is None
        assert game.game_state == GameState.ADVERTISED
        assert game.activity_state == ActivityState.ACTIVE
        assert game.cancelled_reason is None
        assert game.completed_comment is None
        assert game.game_players == {}

    def test_to_advertised_game(self):
        game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["a", "b"])
        advertised = game.to_advertised_game()
        assert advertised.game_id == "game_id"
        assert advertised.name == "name"
        assert advertised.mode == GameMode.STANDARD
        assert advertised.advertiser_handle == "handle"
        assert advertised.players == 3
        assert advertised.available == 3
        assert advertised.visibility == Visibility.PUBLIC
        assert advertised.invited_handles == ["a", "b"]

    def test_get_game_players(self):
        game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["a", "b"])
        gp1 = MagicMock()
        gp2 = MagicMock()
        game.game_players = {"gp1": gp1, "gp2": gp2}
        assert game.get_game_players() == [gp1, gp2]

    def test_get_available_players(self):
        game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["a", "b"])
        gp1 = MagicMock(player_state=PlayerState.WAITING)
        gp2 = MagicMock(player_state=PlayerState.JOINED)
        gp3 = MagicMock(player_state=PlayerState.PLAYING)
        gp4 = MagicMock(player_state=PlayerState.FINISHED)
        gp5 = MagicMock(player_state=PlayerState.QUIT)
        gp6 = MagicMock(player_state=PlayerState.DISCONNECTED)
        game.game_players = {
            "gp1": gp1,
            "gp2": gp2,
            "gp3": gp3,
            "gp4": gp4,
            "gp5": gp5,
            "gp6": gp6,
        }  # obviously not realistic
        assert game.get_available_players() == [
            gp1,
            gp2,
            gp3,
            gp4,
        ]
        assert game.get_available_player_count() == 4

    def test_get_next_turn(self):
        player = MagicMock(player_type=PlayerType.PROGRAMMATIC)
        game = create_test_game()
        game.game_players["handle"] = player
        game._engine.get_next_turn.return_value = "handle"
        assert game.get_next_turn() == ("handle", PlayerType.PROGRAMMATIC)

    def test_get_legal_moves(self):
        move = MagicMock()
        game = create_test_game()
        game._engine.get_legal_moves.return_value = [move]
        assert game.get_legal_moves("handle") == [move]
        game._engine.get_legal_moves.assert_called_once_with("handle")

    def test_get_player_view(self):
        view = MagicMock()
        game = create_test_game()
        game._engine.get_player_view.return_value = view
        assert game.get_player_view("handle") == view
        game._engine.get_player_view.assert_called_once_with("handle")

    def test_get_recent_history(self):
        history = MagicMock()
        game = create_test_game()
        game._engine.get_recent_history.return_value = [history]
        assert game.get_recent_history(10) == [history]
        game._engine.get_recent_history.assert_called_once_with(10)

    def test_is_available_public(self):
        game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["bender", "fry"])

        # If a public game is advertised, then it is available anyone (regardless of whether they are invited)
        game.game_state = GameState.ADVERTISED
        assert game.is_available("leela") is True
        assert game.is_available("bender") is True
        assert game.is_available("fry") is True

        # If any game is not advertised, then it is not available to anyone
        for state in [state for state in GameState if state != GameState.ADVERTISED]:
            game.game_state = state
            assert game.is_available("leela") is False
            assert game.is_available("bender") is False
            assert game.is_available("fry") is False

    def test_is_available_private(self):
        game = TrackedGame("game_id", "handle", "name", GameMode.STANDARD, 3, Visibility.PRIVATE, ["bender", "fry"])

        # If a private game is advertised, then it is available only for an invited handle
        game.game_state = GameState.ADVERTISED
        assert game.is_available("leela") is False
        assert game.is_available("bender") is True
        assert game.is_available("fry") is True

        # If any game is not advertised, then it is not available to anyone
        for state in [state for state in GameState if state != GameState.ADVERTISED]:
            game.game_state = state
            assert game.is_available("leela") is False
            assert game.is_available("bender") is False
            assert game.is_available("fry") is False

    def test_is_advertised(self):
        game = create_test_game()

        game.game_state = GameState.ADVERTISED
        assert game.is_advertised() is True

        for state in [state for state in GameState if state != GameState.ADVERTISED]:
            game.game_state = state
            assert game.is_advertised() is False

    def test_is_playing(self):
        game = create_test_game()

        game.game_state = GameState.PLAYING
        assert game.is_playing() is True

        for state in [state for state in GameState if state != GameState.PLAYING]:
            game.game_state = state
            assert game.is_playing() is False

    def test_is_in_progress(self):
        assert check_is_in_progress(True, True) is True
        assert check_is_in_progress(True, False) is True
        assert check_is_in_progress(False, True) is True
        assert check_is_in_progress(False, False) is False

    def test_is_fully_joined(self):
        assert check_fully_joined(2, 0) is False
        assert check_fully_joined(2, 1) is False
        assert check_fully_joined(2, 2) is True
        assert check_fully_joined(3, 0) is False
        assert check_fully_joined(3, 1) is False
        assert check_fully_joined(3, 2) is False
        assert check_fully_joined(3, 3) is True
        assert check_fully_joined(3, 0) is False
        assert check_fully_joined(4, 1) is False
        assert check_fully_joined(4, 2) is False
        assert check_fully_joined(4, 3) is False
        assert check_fully_joined(4, 4) is True

    def test_is_viable(self):
        for available in [0, 1, 2, 3, 4]:
            assert check_is_viable(True, available) is True
        for available in [0, 1]:
            assert check_is_viable(False, available) is False
        for available in [2, 3, 4]:
            assert check_is_viable(False, available) is True

    def test_is_move_pending(self):
        game = create_test_game()
        game._engine.is_move_pending.return_value = True
        assert game.is_move_pending("handle") is True
        game._engine.is_move_pending.assert_called_once_with("handle")

    def test_is_legal_move(self):
        game = create_test_game()
        game._engine.is_legal_move.return_value = True
        assert game.is_legal_move("handle", "move_id") is True
        game._engine.is_legal_move.assert_called_once_with("handle", "move_id")

    # noinspection PyTypeChecker
    def test_mark_active(self):
        now = pendulum.now()
        game = create_test_game()
        game.last_active_date = None
        game.activity_state = None
        game.mark_active()
        assert game.last_active_date >= now
        assert game.activity_state == ActivityState.ACTIVE

    def test_mark_idle(self):
        game = create_test_game()
        game.activity_state = None
        game.mark_idle()
        assert game.activity_state == ActivityState.IDLE

    def test_mark_inactive(self):
        game = create_test_game()
        game.activity_state = None
        game.mark_inactive()
        assert game.activity_state == ActivityState.INACTIVE

    def test_mark_joined_illegal_state(self):
        game = create_test_game()
        for game_state in [game_state for game_state in GameState if game_state != GameState.ADVERTISED]:
            with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
                game.game_state = game_state
                game.mark_joined("handle")
                assert "leela" not in game.game_players

    def test_mark_joined(self):
        game = create_test_game()
        game.game_state = GameState.ADVERTISED  # otherwise it's an illegal state
        game.mark_joined("leela")
        assert len(game.game_players) == 1
        assert game.game_players["leela"].handle == "leela"
        assert game.game_players["leela"].player_color is None
        assert game.game_players["leela"].player_type == PlayerType.HUMAN
        assert game.game_players["leela"].player_state == PlayerState.JOINED

    def test_mark_started_illegal_state(self):
        game = create_test_game()
        game.players = 4
        game.mark_joined("leela")
        game.mark_joined("bender")
        for game_state in [game_state for game_state in GameState if game_state != GameState.ADVERTISED]:
            with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
                game.game_state = game_state
                game.mark_started()
            assert game.game_state == game_state  # should not change

    # noinspection PyTypeChecker
    def test_mark_started(self):
        now = pendulum.now()
        game = create_test_game()
        game.players = 4
        game.mark_joined("leela")
        game.mark_joined("bender")

        # this stubs out the start_game() method and gives us access to what it returns, for later validation
        # it's a little funky, but works pretty well
        colors = {}

        def stubbed_start_game(mode, handles):
            nonlocal colors
            assert mode == game.mode
            colors = {
                handles[0]: PlayerColor.YELLOW,
                handles[1]: PlayerColor.RED,
                handles[2]: PlayerColor.GREEN,
                handles[3]: PlayerColor.BLUE,
            }
            return colors

        game._engine.start_game = stubbed_start_game  # pylint: disable=assigning-non-slot:

        game.game_state = GameState.ADVERTISED  # otherwise it's an illegal state
        game.last_active_date = None
        game.started_date = None
        game.mark_started()
        assert game.game_state == GameState.PLAYING
        assert game.last_active_date >= now
        assert game.started_date >= now
        assert len(game.game_players) == 4  # the two we added and two programmatic ones

        programmatic = [handle for handle in game.game_players.keys() if handle not in ("leela", "bender")]
        assert len(programmatic) == 2
        for handle in programmatic:
            assert handle in _NAMES  # the name gets randomly assigned
            assert game.game_players[handle].handle == handle
            assert game.game_players[handle].player_color == colors[handle]  # check that it gets what was returned by start_game()
            assert game.game_players[handle].player_type == PlayerType.PROGRAMMATIC
            assert game.game_players[handle].player_state == PlayerState.PLAYING

        human = [handle for handle in game.game_players.keys() if handle in ("leela", "bender")]
        assert len(human) == 2
        for handle in ["leela", "bender"]:
            assert game.game_players[handle].handle == handle
            assert game.game_players[handle].player_color == colors[handle]  # check that it gets what was returned by start_game()
            assert game.game_players[handle].player_type == PlayerType.HUMAN
            assert game.game_players[handle].player_state == PlayerState.PLAYING

    def test_mark_completed_illegal_state(self):
        game = create_test_game()
        for game_state in [game_state for game_state in GameState if game_state != GameState.PLAYING]:
            with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
                game.game_state = game_state
                game.mark_completed("comment")
            assert game.game_state == game_state  # should not change
            game._engine.stop_game.assert_not_called()

    # noinspection PyTypeChecker
    def test_mark_completed(self):
        gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
        gp2 = GamePlayer("gp2", PlayerColor.RED, PlayerType.HUMAN, PlayerState.PLAYING)
        gp1_copy = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.FINISHED)
        gp2_copy = GamePlayer("gp2", PlayerColor.RED, PlayerType.HUMAN, PlayerState.FINISHED)
        now = pendulum.now()
        game = create_test_game()
        game.game_players = {"gp1": gp1, "gp2": gp2}
        game.completed_date = None
        game.game_state = GameState.PLAYING  # otherwise it's an illegal state
        game.completed_comment = None
        game.mark_completed("comment")
        assert game.completed_date >= now
        assert game.game_state == GameState.COMPLETED
        assert game.completed_comment == "comment"
        assert game.game_players == {"gp1": gp1_copy, "gp2": gp2_copy}
        game._engine.stop_game.assert_called_once()

    def test_mark_cancelled_illegal_state(self):
        game = create_test_game()
        for game_state in [game_state for game_state in GameState if game_state not in (GameState.PLAYING, GameState.ADVERTISED)]:
            with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
                game.game_state = game_state
                game.mark_cancelled(CancelledReason.CANCELLED, "comment")
            assert game.game_state == game_state  # should not change
            game._engine.stop_game.assert_not_called()

    # noinspection PyTypeChecker
    def test_mark_cancelled(self):
        for game_state in (GameState.PLAYING, GameState.ADVERTISED):  # either state is legal
            gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
            gp2 = GamePlayer("gp2", PlayerColor.RED, PlayerType.HUMAN, PlayerState.PLAYING)
            gp1_copy = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.FINISHED)
            gp2_copy = GamePlayer("gp2", PlayerColor.RED, PlayerType.HUMAN, PlayerState.FINISHED)
            now = pendulum.now()
            game = create_test_game()
            game.game_players = {"gp1": gp1, "gp2": gp2}
            game.completed_date = None
            game.game_state = game_state
            game.completed_comment = None
            game.mark_cancelled(CancelledReason.NOT_VIABLE, "comment")
            assert game.completed_date >= now
            assert game.game_state == GameState.CANCELLED
            assert game.cancelled_reason == CancelledReason.NOT_VIABLE
            assert game.completed_comment == "comment"
            assert game.game_players == {"gp1": gp1_copy, "gp2": gp2_copy}
            game._engine.stop_game.assert_called_once()

    def test_mark_quit_illegal_state(self):
        for game_state in [game_state for game_state in GameState if game_state not in (GameState.PLAYING, GameState.ADVERTISED)]:
            gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
            game = create_test_game()
            game.game_players = {"gp1": gp1}
            game.game_state = game_state
            with pytest.raises(ProcessingError, match=r"Illegal state for operation"):
                game.mark_quit("gp1")
            assert game.game_state == game_state  # should not change
            assert "gp1" in game.game_players  # should not change

    def test_mark_quit_advertised(self):
        gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
        game = create_test_game()
        game.game_players = {"gp1": gp1}
        game.game_state = GameState.ADVERTISED
        game.mark_quit("gp1")
        assert "gp1" not in game.game_players

    def test_mark_quit_started(self):
        gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
        gp1_copy = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.QUIT)
        game = create_test_game()
        game.game_players = {"gp1": gp1}
        game.game_state = GameState.PLAYING
        game.mark_quit("gp1")
        assert game.game_players["gp1"] == gp1_copy

    def test_execute_move(self):
        game = create_test_game()
        game.execute_move("handle", "move_id")
        game._engine.execute_move.assert_called_once_with("handle", "move_id")

    def test_mark_joined_programmatic(self):
        game = create_test_game()
        game.game_players = {}
        game._mark_joined_programmatic()
        assert len(game.game_players) == 1
        gp1 = list(game.game_players.values())[0]
        assert gp1.handle in _NAMES
        assert gp1.player_color is None  # will be assigned later
        assert gp1.player_type == PlayerType.PROGRAMMATIC
        assert gp1.player_state == PlayerState.JOINED

    def test_assign_handle(self):
        game = create_test_game()

        game.game_players = {}
        handle = game._assign_handle()
        assert handle in _NAMES

        game.game_players = {"gp1": MagicMock(handle="Frodo"), "gp2": MagicMock(handle="Gimli")}
        for _ in range(0, 100):
            handle = game._assign_handle()
            assert handle in _NAMES and handle not in ("Frodo", "Gimli")


class TestStateManager:
    """
    Test the StateManager class.
    """

    def test_mark_active_no_websocket(self):
        player = MagicMock(websocket=None)
        mgr = StateManager()
        mgr.mark_active(player)
        player.mark_active.assert_called_once()

    def test_mark_active_unknown_websocket(self):
        websocket = MagicMock()
        player = MagicMock(websocket=websocket)
        mgr = StateManager()
        with pytest.raises(ProcessingError, match=r"Did not find player websocket"):
            mgr.mark_active(player)
        player.mark_active.assert_not_called()

    def test_mark_active(self):
        websocket = MagicMock()
        player = MagicMock(websocket=websocket)
        mgr = StateManager()
        mgr._websocket_map[websocket] = MagicMock()
        mgr.mark_active(player)
        player.mark_active.assert_called_once()
        mgr._websocket_map[websocket].mark_active.assert_called_once()

    def test_track_websocket_duplicate(self):
        original = MagicMock()
        websocket = MagicMock()
        mgr = StateManager()
        mgr._websocket_map[websocket] = original
        with pytest.raises(ProcessingError, match=r"Duplicate websocket encountered"):
            mgr.track_websocket(websocket)
        assert mgr._websocket_map[websocket] is original

    @patch("apologiesserver.manager.pendulum.now")
    def test_track_websocket(self, now):
        now.return_value = to_date("2020-05-11T16:57:00,000")
        websocket = MagicMock()
        mgr = StateManager()
        mgr.track_websocket(websocket)
        assert mgr._websocket_map[websocket] == TrackedWebsocket(websocket)

    def test_delete_websocket_unknown(self):
        websocket = MagicMock()
        mgr = StateManager()
        mgr.delete_websocket(websocket)  # no-op because it's unknown

    def test_delete_websocket(self):
        websocket = MagicMock()
        mgr = StateManager()
        mgr._websocket_map[websocket] = MagicMock()
        mgr.delete_websocket(websocket)
        assert websocket not in mgr._websocket_map

    def test_get_websocket_count(self):
        mgr = StateManager()
        mgr.track_websocket(MagicMock())
        mgr.track_websocket(MagicMock())
        mgr.track_websocket(MagicMock())
        assert mgr.get_websocket_count() == 3

    def test_lookup_all_websockets(self):
        websocket1 = MagicMock()
        websocket2 = MagicMock()
        websocket3 = MagicMock()
        mgr = StateManager()
        mgr.track_websocket(websocket1)
        mgr.track_websocket(websocket2)
        mgr.track_websocket(websocket3)
        assert mgr.lookup_all_websockets() == [websocket1, websocket2, websocket3]

    def test_lookup_players_for_websocket(self):
        player = MagicMock(player_id="player_id")
        websocket = MagicMock()
        mgr = StateManager()
        mgr.lookup_player = MagicMock()
        mgr.lookup_player.side_effect = [player, None]
        mgr.track_websocket(websocket)
        assert mgr.lookup_players_for_websocket(websocket) == []
        mgr._websocket_map[websocket].player_ids.add("player_id")
        mgr._websocket_map[websocket].player_ids.add("bogus")
        assert mgr.lookup_players_for_websocket(websocket) == [player]

    @patch("apologiesserver.manager.pendulum.now")
    @patch("apologiesserver.manager.uuid4")
    def test_track_game(self, uuid4, now):
        uuid4.return_value = "game_id"
        now.return_value = to_date("2020-05-11T16:57:00,000")
        player = MagicMock(handle="leela")
        advertised = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["a", "b"])
        mgr = StateManager()
        game = mgr.track_game(player, advertised)
        assert game == TrackedGame.for_context("leela", "game_id", advertised)
        assert mgr._game_map["game_id"] is game

    def test_delete_game(self):
        game = MagicMock()
        mgr = StateManager()
        mgr._game_map["game"] = game
        mgr.delete_game(MagicMock(game_id="bogus"))  # safe to delete unknown game
        mgr.delete_game(MagicMock(game_id="game"))
        assert "game" not in mgr._game_map

    def test_get_total_game_count(self):
        game1 = MagicMock()
        mgr = StateManager()
        assert mgr.get_total_game_count() == 0
        mgr._game_map["game1"] = game1
        assert mgr.get_total_game_count() == 1

    def test_get_in_progress_game_count(self):
        game1 = MagicMock()
        game1.is_in_progress.return_value = True
        game2 = MagicMock()
        game2.is_in_progress.return_value = False
        game3 = MagicMock()
        game3.is_in_progress.return_value = True
        mgr = StateManager()
        assert mgr.get_in_progress_game_count() == 0
        mgr._game_map["game1"] = game1
        mgr._game_map["game2"] = game2
        mgr._game_map["game3"] = game3
        assert mgr.get_in_progress_game_count() == 2  # game1 and game3

    def test_lookup_game_game_id(self):
        game = MagicMock()
        mgr = StateManager()
        mgr._game_map["game"] = game
        assert mgr.lookup_game(game_id=None) is None
        assert mgr.lookup_game(game_id="bogus") is None
        assert mgr.lookup_game(game_id="game") is game

    def test_lookup_game_player(self):
        game = MagicMock()
        mgr = StateManager()
        mgr._game_map["game"] = game
        assert mgr.lookup_game(player=MagicMock(game_id=None)) is None
        assert mgr.lookup_game(player=MagicMock(game_id="bogus")) is None
        assert mgr.lookup_game(player=MagicMock(game_id="game")) is game

    def test_lookup_all_games(self):
        game1 = MagicMock()
        game2 = MagicMock()
        game3 = MagicMock()
        mgr = StateManager()
        mgr._game_map["game1"] = game1
        mgr._game_map["game2"] = game2
        mgr._game_map["game3"] = game3
        assert mgr.lookup_all_games() == [game1, game2, game3]

    def test_lookup_in_progress_games(self):
        game1 = MagicMock()
        game1.is_in_progress.return_value = True
        game2 = MagicMock()
        game2.is_in_progress.return_value = False
        game3 = MagicMock()
        game3.is_in_progress.return_value = True
        mgr = StateManager()
        mgr._game_map["game1"] = game1
        mgr._game_map["game2"] = game2
        mgr._game_map["game3"] = game3
        assert mgr.lookup_in_progress_games() == [game1, game3]

    def test_lookup_game_players(self):
        player = MagicMock()
        gp1 = GamePlayer("gp1", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
        gp2 = GamePlayer("gp2", PlayerColor.RED, PlayerType.HUMAN, PlayerState.PLAYING)
        game = create_test_game()
        game.game_players = {"gp1": gp1, "gp2": gp2}
        mgr = StateManager()
        mgr.lookup_player = MagicMock()
        mgr.lookup_player.return_value = player
        assert mgr.lookup_game_players(game) == [player]
        mgr.lookup_player.assert_called_once_with(handle="gp2")

    def test_lookup_available_games(self):
        player = MagicMock(handle="handle")
        game1 = MagicMock()
        game1.is_available.return_value = True
        game2 = MagicMock()
        game2.is_available.return_value = False
        game3 = MagicMock()
        game3.is_available.return_value = True
        mgr = StateManager()
        mgr._game_map["game1"] = game1
        mgr._game_map["game2"] = game2
        mgr._game_map["game3"] = game3
        assert mgr.lookup_available_games(player) == [game1, game3]
        game1.is_available.assert_called_once_with("handle")
        game2.is_available.assert_called_once_with("handle")
        game3.is_available.assert_called_once_with("handle")

    @patch("apologiesserver.manager.pendulum.now")
    @patch("apologiesserver.manager.uuid4")
    def test_track_player(self, uuid4, now):
        uuid4.return_value = "player_id"
        now.return_value = to_date("2020-05-11T16:57:00,000")
        websocket = MagicMock()
        mgr = StateManager()
        mgr.track_websocket(websocket)  # this would always be done ahead of time, so do it here
        mgr._websocket_map[websocket].last_active_date = None  # so we can tell it was reset
        player = mgr.track_player(websocket, "handle")
        assert player == TrackedPlayer.for_context("player_id", websocket, "handle")
        assert mgr._websocket_map[websocket].last_active_date == to_date("2020-05-11T16:57:00,000")  # we always mark the websocket
        assert "player_id" in mgr._websocket_map[websocket].player_ids
        assert mgr._player_map["player_id"] is player
        assert mgr._handle_map["handle"] == "player_id"
        with pytest.raises(ProcessingError, match=r"Handle is already in use"):
            mgr.track_player(websocket, "handle")
        assert mgr._player_map["player_id"] is player
        assert mgr._handle_map["handle"] == "player_id"

    @patch("apologiesserver.manager.pendulum.now")
    def test_retrack_player(self, now):
        now.return_value = to_date("2020-05-11T16:57:00,000")
        player = MagicMock(player_id="player_id")
        websocket = MagicMock()
        mgr = StateManager()
        mgr.track_websocket(websocket)  # this would always be done ahead of time, so do it here
        mgr._websocket_map[websocket].last_active_date = None  # so we can tell it was reset
        mgr.retrack_player(player, websocket)
        assert player.websocket == websocket
        assert mgr._websocket_map[websocket].last_active_date == to_date("2020-05-11T16:57:00,000")  # we always mark the websocket
        assert "player_id" in mgr._websocket_map[websocket].player_ids

    def test_delete_player_no_websocket(self):
        mgr = StateManager()
        mgr._player_map["player_id"] = MagicMock()
        mgr._handle_map["handle"] = MagicMock()
        mgr.delete_player(MagicMock(player_id="bogus", handle="bogus", websocket=MagicMock()))  # safe to delete unknown stuff
        mgr.delete_player(MagicMock(player_id="player_id", handle="handle", websocket=None))
        assert "player_id" not in mgr._player_map
        assert "handle" not in mgr._handle_map

    def test_delete_player_unknown_websocket(self):
        websocket = MagicMock()
        mgr = StateManager()
        # note: not tracking the websocket, so it will not exist when we try to look it up
        mgr._player_map["player_id"] = MagicMock()
        mgr._handle_map["handle"] = MagicMock()
        mgr.delete_player(MagicMock(player_id="bogus", handle="bogus", websocket=MagicMock()))  # safe to delete unknown stuff
        mgr.delete_player(MagicMock(player_id="player_id", handle="handle", websocket=websocket))
        assert "player_id" not in mgr._player_map
        assert "handle" not in mgr._handle_map

    def test_delete_player(self):
        websocket = MagicMock()
        mgr = StateManager()
        mgr.track_websocket(websocket)  # this would always be done ahead of time, so do it here
        mgr._websocket_map[websocket].player_ids.add("player_id")
        mgr._player_map["player_id"] = MagicMock()
        mgr._handle_map["handle"] = MagicMock()
        mgr.delete_player(MagicMock(player_id="bogus", handle="bogus", websocket=MagicMock()))  # safe to delete unknown stuff
        mgr.delete_player(MagicMock(player_id="player_id", handle="handle", websocket=websocket))
        assert "player_id" not in list(mgr._websocket_map[websocket].player_ids)
        assert "player_id" not in mgr._player_map
        assert "handle" not in mgr._handle_map

    def test_get_registered_player_count(self):
        player = MagicMock()
        mgr = StateManager()
        assert mgr.get_registered_player_count() == 0
        mgr._player_map["player"] = player
        assert mgr.get_registered_player_count() == 1

    def test_lookup_player_player_id(self):
        player = MagicMock()
        mgr = StateManager()
        mgr._player_map["player"] = player
        assert mgr.lookup_player(player_id=None) is None
        assert mgr.lookup_player(player_id="bogus") is None
        assert mgr.lookup_player(player_id="player") is player

    def test_lookup_player_handle(self):
        player = MagicMock(player_id="player")
        mgr = StateManager()
        mgr._player_map["player"] = player
        mgr._handle_map["handle"] = "player"
        assert mgr.lookup_player(handle=None) is None
        assert mgr.lookup_player(handle="bogus") is None
        assert mgr.lookup_player(handle="handle") is player

    def test_lookup_all_players(self):
        player = MagicMock()
        mgr = StateManager()
        mgr._player_map["player"] = player
        assert mgr.lookup_all_players() == [player]

    def test_lookup_websocket_activity(self):
        date = to_date("2020-05-12T16:26:00,000")
        websocket = MagicMock()
        tracked = MagicMock(websocket=websocket, last_active_date=date, player_ids=OrderedSet(["1", "2", "3"]))
        mgr = StateManager()
        mgr._websocket_map[websocket] = tracked
        assert mgr.lookup_websocket_activity() == [(tracked, date, 3)]

    def test_lookup_player_activity(self):
        date = to_date("2020-05-12T16:26:00,000")
        tracked = MagicMock(player_id="player_id", last_active_date=date, connection_state=ConnectionState.CONNECTED)
        mgr = StateManager()
        mgr._player_map["player_id"] = tracked
        assert mgr.lookup_player_activity() == [(tracked, date, ConnectionState.CONNECTED)]

    def test_lookup_game_activity(self):
        date = to_date("2020-05-12T16:26:00,000")
        tracked1 = MagicMock(game_id="1", last_active_date=date, completed=False)
        tracked2 = MagicMock(game_id="2", last_active_date=None, completed=True)
        mgr = StateManager()
        mgr._game_map["1"] = tracked1
        mgr._game_map["2"] = tracked2
        assert mgr.lookup_game_activity() == [(tracked1, date)]

    def test_lookup_game_completion(self):
        date = to_date("2020-05-12T16:26:00,000")
        tracked1 = MagicMock(game_id="1", completed_date=None, completed=False)
        tracked2 = MagicMock(game_id="2", completed_date=date, completed=True)
        mgr = StateManager()
        mgr._game_map["1"] = tracked1
        mgr._game_map["2"] = tracked2
        assert mgr.lookup_game_completion() == [(tracked2, date)]


class TestGame:
    """
    Test a complete game via the StateManager interface.

    This basically covers the steps that would happen in a real game via the
    public interface, except directly against the manager interface instead.
    Unlike the public interface, the manager interface is synchronous, which
    makes everything a lot easier to deal with if debugging is required.
    """

    def test_complete_game_standard(self):
        TestGame.play_game(mode=GameMode.STANDARD, player_count=4, handles=["leela"])

    def test_complete_game_adult(self):
        TestGame.play_game(mode=GameMode.ADULT, player_count=3, handles=["leela"])

    # pylint: disable=too-many-locals

    @staticmethod
    def play_game(mode: GameMode, player_count: int, handles: List[str]) -> None:
        assert len(handles) >= 1

        print("Testing complete game with mode %s and 'human' players %s" % (mode, handles))

        websocket = MagicMock()

        name = "Demo Game"
        visibility = Visibility.PUBLIC
        invited_handles: List[str] = []
        advertised = AdvertiseGameContext(name, mode, player_count, visibility, invited_handles)

        mgr = StateManager()
        print("Created state manager.")

        mgr.track_websocket(websocket)
        print("Tracked simulated websocket")

        players: Dict[str, TrackedPlayer] = {}
        for handle in handles:
            player = mgr.track_player(websocket, handle)
            players[handle] = player
            print("Created player %s with id %s" % (player.handle, player.player_id))

        game = mgr.track_game(players[handles[0]], advertised)
        print("Advertised game '%s' with id %s" % (game.name, game.game_id))

        for handle in handles:
            game.mark_joined(handle)
        game.mark_started()
        for player in players.values():
            player.mark_playing()
        print("Started game")
        print(
            "Players: %s"
            % ["%s (%s)" % (player.handle, player.player_color.value) for player in game.get_game_players()]  # type: ignore
        )

        completed = False
        while not completed:
            handle, player_type = game.get_next_turn()
            if player_type == PlayerType.PROGRAMMATIC:
                view = game.get_player_view(handle)
                moves = game.get_legal_moves(handle)
                move = RewardV1InputSource().choose_move(game.mode, view, moves, Rules.evaluate_move)
                (completed, _winner, comment) = game.execute_move(handle, move.id)
            else:
                moves = game.get_legal_moves(handle)
                context = GamePlayerTurnContext.for_moves(handle=handle, game_id=game.game_id, moves=moves)
                move_id = random.choice(list(context.moves.keys()))  # simulates input from the websocket client
                (completed, _winner, comment) = game.execute_move(handle, move_id)
            history = game.get_recent_history(1)
            if history:
                color = "General" if not history[0].color else history[0].color.value
                action = history[0].action
                print("%s - %s" % (color, action))
            if completed:
                for player in players.values():
                    player.mark_quit()
                game.mark_completed(comment)
                print("%s" % comment)

        for player in players.values():
            mgr.delete_player(player)
        mgr.delete_game(game)
        mgr.delete_websocket(websocket)

        print("Test completed")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

import sys
from unittest.mock import Mock, patch

import pytest

from apologiesserver.cli import _example, _lookup_method, cli, run_demo, run_server


class TestUtil:
    """
    Unit tests for the general CLI code.
    """

    def test_cli(self):
        # When a script is invoked, it gets arguments like shown below - that's just how it works
        with patch.object(sys, "argv", ["src/scripts/example", "a", "b", "c"]):
            assert cli("_example") == ["a", "b", "c"]

    def test_lookup_method(self):
        assert _lookup_method("_example") is _example
        with pytest.raises(AttributeError):
            assert _lookup_method("")
        with pytest.raises(AttributeError):
            assert _lookup_method("bogus")


@patch("apologiesserver.cli.server")
@patch("apologiesserver.cli.config")
@patch("apologiesserver.cli.setup_logging")
@patch("apologiesserver.cli.load_config")
class TestRunServer:
    """
    Unit tests for the server CLI.
    """

    def test_run_server_defaults(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = []
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {})
        setup_logging.assert_called_once_with(False, False, False, "path")
        server.assert_called_once()

    def test_run_server_h(self, _load_config, _setup_logging, config, _server):
        config.return_value = Mock(logfile_path="path")
        argv = ["-h"]
        with pytest.raises(SystemExit):
            run_server(argv=argv)

    def test_run_server_help(self, _load_config, _setup_logging, config, _server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--help"]
        with pytest.raises(SystemExit):
            run_server(argv=argv)

    def test_run_server_quiet(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--quiet"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {})
        setup_logging.assert_called_once_with(True, False, False, "path")
        server.assert_called_once()

    def test_run_server_verbose(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--verbose"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {})
        setup_logging.assert_called_once_with(False, True, False, "path")
        server.assert_called_once()

    def test_run_server_debug(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--debug"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {})
        setup_logging.assert_called_once_with(False, False, True, "path")
        server.assert_called_once()

    def test_run_server_config(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--config", "/path/to/config"]
        run_server(argv=argv)
        load_config.assert_called_once_with("/path/to/config", {})
        setup_logging.assert_called_once_with(False, False, False, "path")
        server.assert_called_once()

    def test_run_server_logfile(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--logfile", "/path/to/log"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {"logfile_path": "/path/to/log"})
        setup_logging.assert_called_once_with(False, False, False, "path")
        server.assert_called_once()

    def test_run_server_override_single(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--override", "one:ONE"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {"one": "ONE"})
        setup_logging.assert_called_once_with(False, False, False, "path")
        server.assert_called_once()

    def test_run_server_override_multiple(self, load_config, setup_logging, config, server):
        config.return_value = Mock(logfile_path="path")
        argv = ["--override", "one:ONE", "--logfile", "/path/to/log", "--override", "two:TWO"]
        run_server(argv=argv)
        load_config.assert_called_once_with(None, {"one": "ONE", "two": "TWO", "logfile_path": "/path/to/log"})
        setup_logging.assert_called_once_with(False, False, False, "path")
        server.assert_called_once()


@patch("apologiesserver.cli.demo")
@patch("apologiesserver.cli.setup_logging")
class TestRunDemo:
    """
    Unit tests for the demo CLI.
    """

    def test_run_demo_defaults(self, setup_logging, demo):
        argv = []
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, False, False, None)
        demo.assert_called_once()

    def test_run_demo_h(self, _setup_logging, _demo):
        argv = ["-h"]
        with pytest.raises(SystemExit):
            run_demo(argv=argv)

    def test_run_demo_help(self, _setup_logging, _demo):
        argv = ["--help"]
        with pytest.raises(SystemExit):
            run_demo(argv=argv)

    def test_run_demo_host(self, setup_logging, demo):
        argv = ["--host", "server"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, False, False, None)
        demo.assert_called_once_with(host="server", port=8080)

    def test_run_demo_port(self, setup_logging, demo):
        argv = ["--port", "9000"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, False, False, None)
        demo.assert_called_once_with(host="localhost", port=9000)

    def test_run_demo_quiet(self, setup_logging, demo):
        argv = ["--quiet"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(True, False, False, None)
        demo.assert_called_once_with(host="localhost", port=8080)

    def test_run_demo_verbose(self, setup_logging, demo):
        argv = ["--verbose"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, True, False, None)
        demo.assert_called_once_with(host="localhost", port=8080)

    def test_run_demo_debug(self, setup_logging, demo):
        argv = ["--debug"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, False, True, None)
        demo.assert_called_once_with(host="localhost", port=8080)

    def test_run_demo_logfile(self, setup_logging, demo):
        argv = ["--logfile", "/path/to/log"]
        run_demo(argv=argv)
        setup_logging.assert_called_once_with(False, False, False, "/path/to/log")
        demo.assert_called_once_with(host="localhost", port=8080)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

from unittest.mock import patch

from apologiesserver.scripts import server


class TestServer:
    @patch("apologiesserver.scripts.cli")
    def test_server(self, cli):
        # this reflects how the script is invoked by the Poetry-generated stub
        server()
        cli.assert_called_once_with("run_server")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name,wildcard-import,too-many-lines,use-implicit-booleaness-not-comparison

from unittest.mock import MagicMock, call

import pytest
from apologies.game import GameMode
from apologies.rules import Rules
from asynctest import CoroutineMock
from asynctest import MagicMock as AsyncMock
from asynctest import patch

from apologiesserver.event import EventHandler, RequestContext, TaskQueue
from apologiesserver.interface import *

from .util import to_date


class TestTaskQueue:
    """
    Test the TaskQueue class.
    """

    def test_disconnect(self):
        socket1 = MagicMock()
        socket2 = MagicMock()

        queue = TaskQueue()
        assert queue.is_empty() is True
        assert queue.disconnects == set()

        queue.disconnect(None)
        assert queue.is_empty() is True
        assert queue.disconnects == set()

        queue.disconnect(socket1)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1}

        queue.disconnect(socket2)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1, socket2}

        queue.disconnect(socket1)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1, socket2}

        queue.clear()
        assert queue.is_empty() is True
        assert queue.disconnects == set()

    def test_message(self):
        socket1 = MagicMock()
        socket2 = MagicMock()
        socket3 = MagicMock()

        message1 = MagicMock()
        message1.to_json.return_value = "json1"

        message2 = MagicMock()
        message2.to_json.return_value = "json2"

        message3 = MagicMock()
        message3.to_json.return_value = "json3"

        player = MagicMock(websocket=socket2)

        queue = TaskQueue()
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=None, players=None)
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=[], players=[])
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=[socket1])
        assert queue.is_empty() is False
        assert queue.messages == [("json1", socket1)]

        queue.message(message2, players=[player])
        assert queue.is_empty() is False
        assert queue.messages == [("json1", socket1), ("json2", socket2)]

        queue.message(message3, websockets=[socket1, socket3], players=[player])
        assert queue.is_empty() is False
        assert queue.messages == [
            ("json1", socket1),
            ("json2", socket2),
            ("json3", socket1),
            ("json3", socket3),
            ("json3", socket2),
        ]

        queue.clear()
        assert queue.is_empty() is True
        assert queue.messages == []

    @pytest.mark.asyncio
    @patch("apologiesserver.event.asyncio")
    async def test_execute_empty(self, stub):
        stub.wait = CoroutineMock()
        queue = TaskQueue()
        await queue.execute()
        stub.wait.assert_not_awaited()  # wait() doesn't accept an empty list, so we just don't call it

    @pytest.mark.asyncio
    @patch("apologiesserver.event.asyncio")
    async def test_execute(self, stub):
        stub.wait = CoroutineMock()

        socket1 = MagicMock()
        socket2 = MagicMock()
        socket3 = MagicMock()

        message1 = MagicMock()
        message1.to_json.return_value = "json1"

        message2 = MagicMock()
        message2.to_json.return_value = "json2"

        queue = TaskQueue()
        queue.disconnect(socket1)
        queue.message(message1, websockets=[socket1])  # will be ignored because socket is disconnected
        queue.message(message2, websockets=[socket2, socket3])
        await queue.execute()

        # unfortunately, I can't find a good way to validate the tasks passed to this
        stub.wait.assert_awaited()


@pytest.mark.filterwarnings("ignore:coroutine 'send' was never awaited")
@pytest.mark.filterwarnings("ignore:coroutine 'close' was never awaited")
class TestEventHandler:
    """
    Test the basic EventHandler functionality.
    """

    def test_context_manager(self):
        manager = MagicMock()
        queue = MagicMock()
        queue.clear = MagicMock()
        handler = EventHandler(manager, queue)
        with handler as result:
            assert handler is result  # proves that the right object is returned
            queue.clear.assert_called_once()  # proves that we clear the queue on entering
            queue.clear = MagicMock()  # clear it so we can easily check for 2nd call
        queue.clear.assert_called_once()  # proves that we clear the cache on exiting

    @pytest.mark.asyncio
    async def test_execute(self):
        manager = MagicMock()
        queue = AsyncMock()
        queue.execute = CoroutineMock()
        handler = EventHandler(manager, queue)
        await handler.execute_tasks()
        queue.execute.assert_awaited_once()


@pytest.mark.filterwarnings("ignore:coroutine 'send' was never awaited")
@pytest.mark.filterwarnings("ignore:coroutine 'close' was never awaited")
class TestTaskMethods:
    """
    Test the task-related methods on EventHandler.
    """

    # pylint: disable=invalid-name,too-many-locals
    @patch("apologiesserver.event.pendulum")
    @patch("apologiesserver.event.config")
    def test_handle_idle_websocket_check_task(self, config, pendulum):
        config.return_value = MagicMock(websocket_idle_thresh_min=10, websocket_inactive_thresh_min=20)
        pendulum.now.return_value = to_date("2020-05-11T10:22:00,000")

        w1 = MagicMock()
        w2 = MagicMock()
        w3 = MagicMock()
        w4 = MagicMock()
        w5 = MagicMock()
        w6 = MagicMock()
        w7 = MagicMock()
        w8 = MagicMock()

        # Results: 1 is active (a1), 2 are idle (a2, a3) and 1 is inactive (a4)
        a1 = (w1, to_date("2020-05-11T10:12:00,001"), 0)  # active (9:59.999), no registered players
        a2 = (w2, to_date("2020-05-11T10:12:00,000"), 0)  # idle (10:00.000), no registered players
        a3 = (w3, to_date("2020-05-11T10:02:00,001"), 0)  # idle (19:59.999), no registered players
        a4 = (w4, to_date("2020-05-11T10:02:00,000"), 0)  # inactive (20:00.000), no registered players
        a5 = (w5, to_date("2020-05-11T10:12:00,001"), 5)  # active (9:59.999), with registered players
        a6 = (w6, to_date("2020-05-11T10:12:00,000"), 1)  # idle (10:00.000), with registered players
        a7 = (w7, to_date("2020-05-11T10:02:00,001"), 2)  # idle (19:59.999), with registered players
        a8 = (w8, to_date("2020-05-11T10:02:00,000"), 20)  # inactive (20:00.000), with registered players
        activity = [
            a1,
            a2,
            a3,
            a4,
            a5,
            a6,
            a7,
            a8,
        ]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_websocket_activity.return_value = activity
        handler.handle_websocket_idle_event = MagicMock()
        handler.handle_websocket_inactive_event = MagicMock()

        assert handler.handle_idle_websocket_check_task() == (2, 1)  # anything with a registered player is ignored

        idle_calls = [call(w2), call(w3)]
        inactive_calls = [call(w4)]

        handler.handle_websocket_idle_event.assert_has_calls(idle_calls)
        handler.handle_websocket_inactive_event.assert_has_calls(inactive_calls)

    # pylint: disable=too-many-locals,invalid-name
    @patch("apologiesserver.event.pendulum")
    @patch("apologiesserver.event.config")
    def test_handle_idle_player_check_task(self, config, pendulum):
        config.return_value = MagicMock(player_idle_thresh_min=10, player_inactive_thresh_min=20)
        pendulum.now.return_value = to_date("2020-05-11T10:22:00,000")

        p1 = MagicMock()
        p2 = MagicMock()
        p3 = MagicMock()
        p4 = MagicMock()
        p5 = MagicMock()
        p6 = MagicMock()
        p7 = MagicMock()
        p8 = MagicMock()

        # Results: 2 are active (a1, a5), 2 are idle (a2, a3) and 4 are inactive (a4, a6, a7, a8)
        # A disconnected player is either active or inactive, it can't be idle
        a1 = (p1, to_date("2020-05-11T10:12:00,001"), ConnectionState.CONNECTED)  # connected and active (9:59.999)
        a2 = (p2, to_date("2020-05-11T10:12:00,000"), ConnectionState.CONNECTED)  # connected and idle (10:00.000)
        a3 = (p3, to_date("2020-05-11T10:02:00,001"), ConnectionState.CONNECTED)  # connected and idle (19:59.999)
        a4 = (p4, to_date("2020-05-11T10:02:00,000"), ConnectionState.CONNECTED)  # connected and inactive (20:00.000)
        a5 = (p5, to_date("2020-05-11T10:12:00,001"), ConnectionState.DISCONNECTED)  # disconnected and active (9:59.999)
        a6 = (p6, to_date("2020-05-11T10:12:00,000"), ConnectionState.DISCONNECTED)  # disconnected and idle (10:00.000)
        a7 = (p7, to_date("2020-05-11T10:12:00,000"), ConnectionState.DISCONNECTED)  # disconnected and idle (19:59.999)
        a8 = (p8, to_date("2020-05-11T10:02:00,000"), ConnectionState.DISCONNECTED)  # disconnected and inactive (20:00.000)
        activity = [a1, a2, a3, a4, a5, a6, a7, a8]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_player_activity.return_value = activity
        handler.handle_player_idle_event = MagicMock()
        handler.handle_player_inactive_event = MagicMock()

        assert handler.handle_idle_player_check_task() == (2, 4)

        idle_calls = [call(p2), call(p3)]
        inactive_calls = [call(p4), call(p6), call(p7), call(p8)]

        handler.handle_player_idle_event.assert_has_calls(idle_calls)
        handler.handle_player_inactive_event.assert_has_calls(inactive_calls)

    # pylint: disable=invalid-name
    @patch("apologiesserver.event.pendulum")
    @patch("apologiesserver.event.config")
    def test_handle_idle_game_check_task(self, config, pendulum):
        config.return_value = MagicMock(game_idle_thresh_min=10, game_inactive_thresh_min=20)
        pendulum.now.return_value = to_date("2020-05-11T10:22:00,000")

        g1 = MagicMock()
        g2 = MagicMock()
        g3 = MagicMock()
        g4 = MagicMock()

        # Results: 1 is active (a1), 2 are idle (a2, a3) and 1 is inactive (a4)
        a1 = (g1, to_date("2020-05-11T10:12:00,001"))  # active (9:59.999)
        a2 = (g2, to_date("2020-05-11T10:12:00,000"))  # idle (10:00.000)
        a3 = (g3, to_date("2020-05-11T10:02:00,001"))  # idle (19:59.999)
        a4 = (g4, to_date("2020-05-11T10:02:00,000"))  # inactive (20:00.000)
        activity = [a1, a2, a3, a4]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_activity.return_value = activity
        handler.handle_game_idle_event = MagicMock()
        handler.handle_game_inactive_event = MagicMock()

        assert handler.handle_idle_game_check_task() == (2, 1)

        idle_calls = [call(g2), call(g3)]
        inactive_calls = [call(g4)]

        handler.handle_game_idle_event.assert_has_calls(idle_calls)
        handler.handle_game_inactive_event.assert_has_calls(inactive_calls)

    # pylint: disable=invalid-name
    @patch("apologiesserver.event.pendulum")
    @patch("apologiesserver.event.config")
    def test_handle_obsolete_game_check_task(self, config, pendulum):
        config.return_value = MagicMock(game_retention_thresh_min=10)
        pendulum.now.return_value = to_date("2020-05-11T10:22:00,000")

        g1 = MagicMock()
        g2 = MagicMock()

        # Results: 1 is young enough to keep (a1) and 1 is obsolete (a2)
        a1 = (g1, to_date("2020-05-11T10:12:00,001"))  # young enough to keep (9:59.999)
        a2 = (g2, to_date("2020-05-11T10:12:00,000"))  # obsolete (10:00.000)
        completion = [a1, a2]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_completion.return_value = completion
        handler.handle_game_obsolete_event = MagicMock()

        assert handler.handle_obsolete_game_check_task() == 1

        obsolete_calls = [call(g2)]

        handler.handle_game_obsolete_event.assert_has_calls(obsolete_calls)


# pylint: disable=too-many-public-methods
@pytest.mark.filterwarnings("ignore:coroutine 'send' was never awaited")
@pytest.mark.filterwarnings("ignore:coroutine 'close' was never awaited")
class TestRequestMethods:
    """
    Test the request-related methods on EventHandler.
    """

    @patch("apologiesserver.event.config")
    def test_handle_register_player_request_below_limit(self, config):
        config.return_value = MagicMock(registered_player_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_registered_player_count.return_value = 4
        handler.handle_player_registered_event = MagicMock()
        websocket = MagicMock()
        context = RegisterPlayerContext(handle="leela")
        message = Message(MessageType.REGISTER_PLAYER, context=context)
        handler.handle_register_player_request(message, websocket)
        assert handler.queue.is_empty()
        handler.handle_player_registered_event.assert_called_once_with(websocket, "leela")

    @patch("apologiesserver.event.config")
    def test_handle_register_player_request_at_limit(self, config):
        config.return_value = MagicMock(registered_player_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_registered_player_count.return_value = 5
        handler.handle_player_registered_event = MagicMock()
        websocket = MagicMock()
        context = RegisterPlayerContext(handle="leela")
        message = Message(MessageType.REGISTER_PLAYER, context=context)
        with pytest.raises(ProcessingError, match=r"System user limit reached"):
            handler.handle_register_player_request(message, websocket)
        assert handler.queue.is_empty()
        handler.handle_player_registered_event.assert_not_called()

    def test_handle_reregister_player_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_reregistered_event = MagicMock()
        context = ReregisterPlayerContext(handle="leela")
        message = Message(MessageType.REREGISTER_PLAYER, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_reregister_player_request(request)
        handler.handle_player_reregistered_event.assert_called_once_with(player, websocket)

    def test_handle_unregister_player_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        message = Message(MessageType.UNREGISTER_PLAYER, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_unregister_player_request(request)
        handler.handle_player_unregistered_event.assert_called_once_with(player, game)

    def test_handle_list_players_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_registered_players_event = MagicMock()
        message = Message(MessageType.LIST_PLAYERS, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_list_players_request(request)
        handler.handle_registered_players_event.assert_called_once_with(player)

    def test_handle_advertise_game_request_already_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is already playing a game"):
            handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_advertise_game_request_below_limit(self, config):
        config.return_value = MagicMock(total_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_total_game_count.return_value = 4
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_called_once_with(player, context)

    @patch("apologiesserver.event.config")
    def test_handle_advertise_game_request_at_limit(self, config):
        config.return_value = MagicMock(total_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_total_game_count.return_value = 5
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"System game limit reached"):
            handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_not_called()

    def test_handle_list_available_games_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_available_games_event = MagicMock()
        message = Message(MessageType.LIST_AVAILABLE_GAMES, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_list_available_games_request(request)
        handler.handle_available_games_event.assert_called_once_with(player)

    def test_handle_join_game_request_already_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_joined_event = MagicMock()
        context = JoinGameContext(game_id="game")
        message = Message(MessageType.JOIN_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is already playing a game"):
            handler.handle_join_game_request(request)
        handler.handle_game_joined_event.assert_not_called()

    def test_handle_join_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_joined_event = MagicMock()
        context = JoinGameContext(game_id="game")
        message = Message(MessageType.JOIN_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        handler.handle_join_game_request(request)
        handler.handle_game_joined_event.assert_called_once_with(player, game_id="game")

    def test_handle_quit_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request_not_in_progress(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_in_progress.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not in progress"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Advertiser may not quit a game"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_called_once_with(player, game)

    def test_handle_start_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_start_game_request_already_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = True
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is already being played"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_start_game_request_not_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="bender")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player did not advertise this game"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_start_game_request_below_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_in_progress_game_count.return_value = 4
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_called_once_with(game)

    @patch("apologiesserver.event.config")
    def test_handle_start_game_request_at_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_in_progress_game_count.return_value = 5
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"System game limit reached"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_cancel_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request_not_in_progress(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_in_progress.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not in progress"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request_not_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="bender")
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player did not advertise this game"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        request = RequestContext(message, websocket, player, game)
        handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.CANCELLED)

    def test_handle_execute_move_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_execute_move_request(request)
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_execute_move_request(request)
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_no_move_pending(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"No move is pending for this player"):
            handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.is_legal_move.assert_not_called()
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_illegal_move(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        game.is_legal_move.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"The chosen move is not legal"):
            handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.is_legal_move.assert_called_once_with("leela", "move")
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        handler.handle_game_player_move_event.assert_called_once_with(player, game, "move")

    def test_handle_optimal_move_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_optimal_move_request(request)
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_optimal_move_request(request)
        game.mark_active.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request_no_move_pending(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"No move is pending for this player"):
            handler.handle_optimal_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.mark_active.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_optimal_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.mark_active.assert_called_once()
        handler.handle_game_programmatic_move_event.assert_called_once_with("leela", game)

    def test_handle_retrieve_game_state_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.assert_not_called()

    def test_handle_retrieve_game_state_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.called_once_with(game, player)

    def test_handle_retrieve_game_state_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.called_once_with(game, player)

    def test_handle_send_message_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_message_received_event = MagicMock()
        context = SendMessageContext(message="hello", recipient_handles=["fry", "bender"])
        message = Message(MessageType.SEND_MESSAGE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_send_message_request(request)
        handler.handle_player_message_received_event.called_once_with("handle", ["fry", "bender"], "hello")


# pylint: disable=assigning-non-slot:
class TestEventMethods:
    """
    Test the event-related methods on EventHandler.
    """

    def test_handle_server_shutdown_event(self):
        websocket = MagicMock()
        game = MagicMock()
        message = Message(MessageType.SERVER_SHUTDOWN)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_all_websockets.return_value = [websocket]
        handler.manager.lookup_in_progress_games.return_value = [game]
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_server_shutdown_event()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket])
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.SHUTDOWN, notify=False)

    @patch("apologiesserver.event.config")
    def test_handle_websocket_connected_event_below_limit(self, config):
        config.return_value = MagicMock(websocket_limit=5)
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.track_websocket = MagicMock()
        handler.manager.get_websocket_count.return_value = 4
        handler.handle_websocket_connected_event(websocket)
        handler.manager.track_websocket.assert_called_once_with(websocket)

    @patch("apologiesserver.event.config")
    def test_handle_websocket_connected_event_at_limit(self, config):
        config.return_value = MagicMock(websocket_limit=5)
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.track_websocket = MagicMock()
        handler.manager.get_websocket_count.return_value = 5
        with pytest.raises(ProcessingError, match=r"Connection limit reached"):
            handler.handle_websocket_connected_event(websocket)
        handler.manager.track_websocket.assert_not_called()

    def test_handle_websocket_disconnected_event(self):
        player = MagicMock()
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_player_disconnected_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.manager.delete_websocket = MagicMock()
        handler.manager.lookup_players_for_websocket.return_value = [player]
        handler.handle_websocket_disconnected_event(websocket)
        handler.handle_player_disconnected_event.assert_called_once_with(player)
        handler.manager.delete_websocket.assert_called_once_with(websocket)

    def test_handle_websocket_idle_event_no_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_websocket_idle_event(websocket)
        websocket.mark_idle.assert_not_called()
        handler.queue.message.assert_not_called()

    def test_handle_websocket_idle_event_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=None)
        message = Message(MessageType.WEBSOCKET_IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_websocket_idle_event(websocket)
        websocket.mark_idle.assert_called_once()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket.websocket])

    def test_handle_websocket_inactive_event_no_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_websocket_inactive_event(websocket)
        websocket.mark_inactive.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.queue.disconnect.assert_not_called()

    def test_handle_websocket_inactive_event_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=None)
        message = Message(MessageType.WEBSOCKET_INACTIVE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_websocket_inactive_event(websocket)
        websocket.mark_inactive.assert_called_once()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket.websocket])
        handler.queue.disconnect.assert_called_once_with(websocket.websocket)

    def test_handle_registered_players_event(self):
        player = MagicMock()
        result = MagicMock()
        registered = MagicMock()
        result.to_registered_player.return_value = registered
        context = RegisteredPlayersContext(players=[registered])
        message = Message(MessageType.REGISTERED_PLAYERS, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_all_players.return_value = [result]
        handler.handle_registered_players_event(player)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_available_games_event(self):
        player = MagicMock()
        advertised = MagicMock()
        game = MagicMock()
        game.to_advertised_game.return_value = advertised
        context = AvailableGamesContext(games=[advertised])
        message = Message(MessageType.AVAILABLE_GAMES, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_available_games = MagicMock()
        handler.manager.lookup_available_games.return_value = [game]
        handler.handle_available_games_event(player)
        handler.manager.lookup_available_games.assert_called_once_with(player)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_player_registered_event(self):
        websocket = MagicMock()
        player = MagicMock(player_id="player_id", handle="handle")
        context = PlayerRegisteredContext(handle="handle")
        message = Message(MessageType.PLAYER_REGISTERED, player_id="player_id", context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.track_player = MagicMock()
        handler.manager.track_player.return_value = player
        handler.handle_player_registered_event(websocket, "leela")
        handler.manager.track_player.assert_called_once_with(websocket, "leela")
        handler.queue.message.assert_called_once_with(message, websockets=[websocket])

    def test_handle_player_reregistered_event(self):
        websocket = MagicMock()
        player = MagicMock(player_id="player_id", handle="handle")
        context = PlayerRegisteredContext(handle="handle")
        message = Message(MessageType.PLAYER_REGISTERED, player_id="player_id", context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.retrack_player = MagicMock()
        handler.handle_player_reregistered_event(player, websocket)
        handler.manager.retrack_player.assert_called_once_with(player, websocket)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_player_unregistered_event_no_game(self):
        player = MagicMock(player_id="player_id", handle="leela")
        context = PlayerUnregisteredContext(handle="leela")
        message = Message(MessageType.PLAYER_UNREGISTERED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.delete_player = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_player_unregistered_event(player)
        player.mark_quit.assert_called_once()
        handler.handle_game_player_left_event.assert_not_called()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.delete_player.assert_called_once_with(player)

    def test_handle_player_unregistered_event_with_game(self):
        comment = "Player leela unregistered"
        player = MagicMock(player_id="player_id", handle="leela")
        context = PlayerUnregisteredContext(handle="leela")
        message = Message(MessageType.PLAYER_UNREGISTERED, context=context)
        game = MagicMock()
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.delete_player = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_player_unregistered_event(player, game)
        player.mark_quit.assert_called_once()
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.delete_player.assert_called_once_with(player)

    def test_handle_player_disconnected_event_no_game(self):
        player = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_game_player_left_event = MagicMock()
        handler.manager.lookup_game = MagicMock()
        handler.manager.lookup_game.return_value = None
        handler.handle_player_disconnected_event(player)
        handler.manager.lookup_game.assert_called_once_with(player=player)
        player.mark_disconnected.assert_called_once()
        handler.handle_game_player_left_event.assert_not_called()

    def test_handle_player_disconnected_event_with_game(self):
        comment = "Player leela disconnected"
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.handle_game_player_left_event = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_player_disconnected_event(player)
        handler.manager.lookup_game.assert_called_once_with(player=player)
        player.mark_disconnected.assert_called_once()
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)

    def test_handle_player_idle_event_no_change(self):
        player = MagicMock(activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_player_idle_event(player)
        handler.queue.message.assert_not_called()
        player.mark_idle.assert_not_called()

    def test_handle_player_idle_event_change(self):
        player = MagicMock(handle="handle", activity_state=None)
        context = PlayerIdleContext(handle="handle")
        message = Message(MessageType.PLAYER_IDLE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_player_idle_event(player)
        handler.queue.message.assert_called_once_with(message, players=[player])
        player.mark_idle.assert_called_once()

    def test_handle_player_inactive_event_no_change(self):
        websocket = MagicMock()
        player = MagicMock(websocket=websocket, activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_player_inactive_event(player)
        player.mark_inactive.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.queue.disconnect.assert_not_called()
        handler.handle_player_unregistered_event.assert_not_called()

    def test_handle_player_inactive_event_change(self):
        websocket = MagicMock()
        player = MagicMock(handle="handle", websocket=websocket, activity_state=None)
        game = MagicMock()
        context = PlayerInactiveContext(handle="handle")
        message = Message(MessageType.PLAYER_INACTIVE, context=context)
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_player_inactive_event(player)
        player.mark_inactive.assert_called_once()
        handler.manager.lookup_game.assert_called_once_with(player=player)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.queue.disconnect.assert_not_called()  # we don't close now, only when the websocket itself goes inactive
        handler.handle_player_unregistered_event.assert_called_once_with(player, game)

    def test_handle_player_message_received_event(self):
        fry = MagicMock()
        bender = MagicMock()
        context = PlayerMessageReceivedContext("leela", ["fry", "bender"], "hello")
        message = Message(MessageType.PLAYER_MESSAGE_RECEIVED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.manager.lookup_player.side_effect = [fry, bender]
        handler.handle_player_message_received_event("leela", ["fry", "bender"], "hello")
        handler.manager.lookup_player.assert_has_calls([call(handle="fry"), call(handle="bender")])
        handler.queue.message.assert_called_once_with(message, players=[fry, bender])

    def test_handle_game_advertised_event(self):
        player = MagicMock()
        advertised = MagicMock()
        game = MagicMock()
        result = MagicMock()
        game.to_advertised_game.return_value = result
        context = GameAdvertisedContext(game=result)
        message = Message(MessageType.GAME_ADVERTISED, context=context)
        handler = EventHandler(MagicMock())
        handler.manager.track_game = MagicMock(return_value=game)
        handler.handle_game_invitation_event = MagicMock()
        handler.handle_game_joined_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.handle_game_advertised_event(player, advertised)
        handler.manager.track_game.assert_called_once_with(player, advertised)
        handler.handle_game_joined_event.assert_called_once_with(player, game=game)
        handler.handle_game_invitation_event.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_game_invitation_event_none_invited(self):
        game = MagicMock(invited_handles=[])
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.handle_game_invitation_event(game)
        handler.manager.lookup_player.assert_not_called()
        handler.queue.message.assert_not_called()

    def test_handle_game_invitation_event_with_invited(self):
        fry = MagicMock()
        bender = MagicMock()
        game = MagicMock(invited_handles=["fry", "bender"])
        advertised = MagicMock()
        game.to_advertised_game.return_value = advertised
        context = GameInvitationContext(advertised)
        message = Message(MessageType.GAME_INVITATION, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.manager.lookup_player.side_effect = [fry, bender]
        handler.handle_game_invitation_event(game)
        handler.manager.lookup_player.assert_has_calls([call(handle="fry"), call(handle="bender")])
        handler.queue.message.assert_called_once_with(message, players=[fry, bender])

    def test_handle_game_joined_event_bad_call(self):
        player = MagicMock()
        handler = EventHandler(MagicMock())
        with pytest.raises(ProcessingError, match=r"Invalid arguments"):
            handler.handle_game_joined_event(player, game_id=None, game=None)  # need to pass either id or game

    def test_handle_game_joined_event_not_found(self):
        player = MagicMock(handle="handle")
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game = MagicMock(return_value=None)
        with pytest.raises(ProcessingError, match=r"Unknown or invalid game"):
            handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")

    def test_handle_game_joined_event_not_available(self):
        player = MagicMock(handle="handle")
        game = MagicMock()
        game.is_available.return_value = False
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game = MagicMock(return_value=game)
        with pytest.raises(ProcessingError, match=r"Unknown or invalid game"):
            handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")
        game.is_available.assert_called_once_with("handle")

    def test_handle_game_joined_event_pending_by_id(self):
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = False
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")
        game.is_available.assert_called_once_with("handle")
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()

    def test_handle_game_joined_event_pending_for_game(self):
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = False
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_game_joined_event_fully_joined(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = True
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.get_in_progress_game_count.return_value = 4
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_called_once_with(game)

    @patch("apologiesserver.event.config")
    def test_handle_game_joined_event_game_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = True
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.get_in_progress_game_count.return_value = 5
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()  # because in-progress limit was reached

    def test_handle_game_started_event(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameStartedContext(game_id="game")
        message = Message(MessageType.GAME_STARTED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_started_event(game)
        game.mark_active.assert_called_once()
        game.mark_started.assert_called_once()
        player.mark_playing.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Game started")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_next_turn_event.assert_called_once_with(game)

    def test_handle_game_cancelled_event_notify(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameCancelledContext("game", CancelledReason.SHUTDOWN, "comment")
        message = Message(MessageType.GAME_CANCELLED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_cancelled_event(game, CancelledReason.SHUTDOWN, "comment", notify=True)
        player.mark_quit.assert_called_once()
        game.mark_cancelled.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_state_change_event.assert_called_once_with(game)

    def test_handle_game_cancelled_event_no_notify(self):
        player = MagicMock()
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_cancelled_event(game, CancelledReason.SHUTDOWN, "comment", notify=False)
        player.mark_quit.assert_called_once()
        game.mark_cancelled.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_not_called()  # notification disabled
        handler.handle_game_state_change_event.assert_not_called()  # notification disabled

    def test_handle_game_completed_event(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameCompletedContext("game", "winner", "comment")
        message = Message(MessageType.GAME_COMPLETED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_completed_event(game, "winner", "comment")
        player.mark_quit.assert_called_once()
        game.mark_completed.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_state_change_event.assert_called_once_with(game)

    def test_handle_game_idle_event_no_change(self):
        game = MagicMock(activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_idle_event(game)
        handler.queue.message.assert_not_called()

    def test_handle_game_idle_event_change(self):
        player = MagicMock()
        game = MagicMock(game_id="game", activity_state=None)
        context = GameIdleContext(game_id="game")
        message = Message(MessageType.GAME_IDLE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_idle_event(game)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_game_inactive_event(self):
        game = MagicMock(activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_inactive_event(game)
        game.mark_inactive.assert_not_called()
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_game_inactive_event_change(self):
        player = MagicMock()
        game = MagicMock(game_id="game", activity_state=None)
        context = GameInactiveContext(game_id="game")
        message = Message(MessageType.GAME_INACTIVE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.manager.lookup_game_players.return_value = [player]
        handler.handle_game_inactive_event(game)
        game.mark_inactive.assert_called_once()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.INACTIVE)

    def test_handle_game_obsolete_event(self):
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.delete_game = MagicMock()
        handler.handle_game_obsolete_event(game)
        handler.manager.delete_game.assert_called_once_with(game)

    def test_handle_game_player_quit_event(self):
        comment = "Player leela quit"
        player = MagicMock(handle="leela")
        game = MagicMock(game_id="game")
        context = GamePlayerQuitContext(handle="leela", game_id="game")
        message = Message(MessageType.GAME_PLAYER_QUIT, context=context)
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_game_player_quit_event(player, game)
        game.mark_active.assert_called_once()
        player.mark_quit.assert_called_once()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)

    def test_handle_game_player_left_event_advertiser(self):
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.CANCELLED, "comment")
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_not_viable(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.NOT_VIABLE, "comment")
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_not_playing(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_playing(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = True
        game.is_move_pending.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_pending(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = True
        game.is_move_pending.return_value = True
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_called_once_with("bender", game)

    def test_handle_game_player_move_event(self):
        player = MagicMock(handle="handle")
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_game_move_event = MagicMock()
        handler.handle_game_player_move_event(player, game, "move")
        handler.handle_game_move_event.assert_called_once_with("handle", game, "move")

    @patch("apologiesserver.event.RewardV1InputSource")
    def test_handle_game_programmatic_move_event(self, obj):
        source = MagicMock()
        obj.return_value = source  # mock the constructor
        view = MagicMock()
        move = MagicMock(id="move")
        game = MagicMock(mode=GameMode.STANDARD)
        game.get_player_view.return_value = view
        game.get_legal_moves.return_value = [move]
        source.choose_move.return_value = move
        handler = EventHandler(MagicMock())
        handler.handle_game_move_event = MagicMock()
        handler.handle_game_programmatic_move_event("handle", game)
        game.get_player_view.assert_called_once_with("handle")
        game.get_legal_moves.assert_called_once_with("handle")
        source.choose_move.assert_called_once_with(GameMode.STANDARD, view, [move], Rules.evaluate_move)
        handler.handle_game_move_event.assert_called_once_with("handle", game, "move")

    def test_handle_game_move_event_completed(self):
        game = MagicMock()
        game.execute_move.return_value = (True, "winner", "comment")
        handler = EventHandler(MagicMock())
        handler.handle_game_completed_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.handle_game_move_event("handle", game, "move_id")
        game.mark_active.assert_called_once()
        game.execute_move.assert_called_once_with("handle", "move_id")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_completed_event.assert_called_once_with(game, "winner", "comment")
        handler.handle_game_next_turn_event.assert_not_called()

    def test_handle_game_player_event_not_completed(self):
        game = MagicMock()
        game.execute_move.return_value = (False, None, None)
        handler = EventHandler(MagicMock())
        handler.handle_game_completed_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.handle_game_move_event("handle", game, "move_id")
        game.mark_active.assert_called_once()
        game.execute_move.assert_called_once_with("handle", "move_id")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_completed_event.assert_not_called()
        handler.handle_game_next_turn_event.assert_called_once_with(game)

    def test_handle_game_next_turn_event_programmatic(self):
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.PROGRAMMATIC)
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
        handler.handle_game_player_turn_event.assert_not_called()

    def test_handle_game_next_turn_event_human_not_found(self):
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
        game.is_playing.return_value = True
        handler = EventHandler(MagicMock())
        handler.manager.lookup_player = MagicMock(return_value=None)  # player is not found
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
        game.get_legal_moves.assert_not_called()
        handler.handle_game_player_turn_event.assert_not_called()
        handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_next_turn_event_human_not_playing(self):
        for player_state in [player_state for player_state in PlayerState if player_state != PlayerState.PLAYING]:
            player = MagicMock(player_state=player_state)
            game = MagicMock()
            game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
            game.is_playing.return_value = True
            handler = EventHandler(MagicMock())
            handler.manager.lookup_player = MagicMock(return_value=player)
            handler.handle_game_programmatic_move_event = MagicMock()
            handler.handle_game_player_turn_event = MagicMock()
            handler.handle_game_next_turn_event(game)
            handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
            game.get_legal_moves.assert_not_called()
            handler.handle_game_player_turn_event.assert_not_called()
            handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_next_turn_event_human_playing(self):
        move = MagicMock()
        player = MagicMock(player_state=PlayerState.PLAYING)
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
        game.is_playing.return_value = True
        game.get_legal_moves.return_value = [move]
        handler = EventHandler(MagicMock())
        handler.manager.lookup_player = MagicMock(return_value=player)
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_not_called()
        game.get_legal_moves.assert_called_once_with("handle")
        handler.handle_game_player_turn_event.assert_called_once_with(player, [move])
        handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_player_change_event(self):
        game_player = MagicMock()
        player = MagicMock()
        game = MagicMock(game_id="game")
        game.get_game_players.return_value = [game_player]
        context = GamePlayerChangeContext(game_id="game", comment="comment", players=[game_player])
        message = Message(MessageType.GAME_PLAYER_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_player_change_event(game, "comment")
        handler.queue.message.assert_called_once_with(message, players=[player])

    @patch("apologiesserver.event.GameStateChangeContext")
    def test_handle_game_state_change_event_specific_player(self, game_state_change_context):
        history = MagicMock()
        context = GameStateChangeContext("game", player=None, opponents=None, recent_history=[history])
        player = MagicMock(handle="handle")
        view = MagicMock()
        game = MagicMock(game_id="game")
        game.get_player_view.return_value = view
        game.get_recent_history.return_value = [history]
        game_state_change_context.for_context.return_value = context
        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock()
        handler.handle_game_state_change_event(game, player)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_called_with("handle")
        game.get_recent_history.assert_called_once_with(10)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.lookup_game_players.assert_not_called()

    def test_handle_game_state_change_event_game_not_playing(self):
        game = MagicMock(game_id="game")
        game.is_playing.return_value = False
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock()
        handler.handle_game_state_change_event(game)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.manager.lookup_game_players.assert_not_called()

    @patch("apologiesserver.event.GameStateChangeContext")
    def test_handle_game_state_change_event_game_players(self, game_state_change_context):
        game = MagicMock(game_id="game", name="name", mode=GameMode.ADULT, advertiser_handle="advertiser")
        history = MagicMock()
        context = GameStateChangeContext("game", player=None, opponents=None, recent_history=[history])
        player = MagicMock(handle="handle")
        view = MagicMock()
        game = MagicMock(game_id="game")
        game.is_playing.return_value = True
        game.get_player_view.return_value = view
        game.get_recent_history.return_value = [history]
        game_state_change_context.for_context.return_value = context
        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.queue.message = MagicMock()
        handler.handle_game_state_change_event(game)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_called_with("handle")
        game.get_recent_history.assert_called_once_with(10)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.lookup_game_players.assert_called_once_with(game)

    @patch("apologiesserver.event.GamePlayerTurnContext")
    def test_handle_game_player_turn_event(self, game_player_turn_context):
        context = GamePlayerTurnContext("handle", "game", None, None)
        player = MagicMock(handle="handle", game_id="game")
        moves = [MagicMock()]
        game_player_turn_context.for_moves.return_value = context
        message = Message(MessageType.GAME_PLAYER_TURN, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_turn_event(player, moves)
        handler.queue.message.assert_called_once_with(message, players=[player])



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=unsubscriptable-object

"""
Shared test utilities.
"""

import random
import string
from unittest.mock import MagicMock

from asynctest import CoroutineMock
from pendulum.datetime import DateTime
from pendulum.parser import parse


def random_string(length: int = 10) -> str:
    chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
    return "".join([random.choice(chars) for _ in range(0, length)])


# noinspection PyTypeChecker
def to_date(date: str) -> DateTime:
    # This function seems to have the wrong type hint
    return parse(date)  # type: ignore


def mock_handler() -> MagicMock:
    """Create a mocked handler that can be locked and used as expected."""
    lock = CoroutineMock()
    handler = MagicMock()
    handler.__enter__.return_value = handler
    handler.execute_tasks = CoroutineMock()
    handler.manager = MagicMock()
    handler.manager.lock = lock
    handler.manager.lock.__aenter__ = lock
    handler.manager.lock.__aexit__ = lock
    return handler



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['apologiesserver']

package_data = \
{'': ['*']}

install_requires = \
['apologies>=0.1.36,<0.2.0',
 'asyncio-periodic>=2019.2,<2020.0',
 'ordered-set>=4.1.0,<5.0.0',
 'websockets>=10.3,<11.0']

entry_points = \
{'console_scripts': ['apologies-server = apologiesserver.scripts:server']}

setup_kwargs = {
    'name': 'apologiesserver',
    'version': '0.1.23',
    'description': 'Websocket server to interactively play the Apologies game',
    'long_description': '# Apologies Server\n\n[![pypi](https://img.shields.io/pypi/v/apologiesserver.svg)](https://pypi.org/project/apologiesserver/)\n[![license](https://img.shields.io/pypi/l/apologiesserver.svg)](https://github.com/pronovic/apologies-server/blob/master/LICENSE)\n[![wheel](https://img.shields.io/pypi/wheel/apologiesserver.svg)](https://pypi.org/project/apologiesserver/)\n[![python](https://img.shields.io/pypi/pyversions/apologiesserver.svg)](https://pypi.org/project/apologiesserver/)\n[![Test Suite](https://github.com/pronovic/apologies-server/workflows/Test%20Suite/badge.svg)](https://github.com/pronovic/apologies-server/actions?query=workflow%3A%22Test+Suite%22)\n[![docs](https://readthedocs.org/projects/apologies-server/badge/?version=stable&style=flat)](https://apologies-server.readthedocs.io/en/stable/)\n[![coverage](https://coveralls.io/repos/github/pronovic/apologies-server/badge.svg?branch=master)](https://coveralls.io/github/pronovic/apologies-server?branch=master)\n\n[Apologies Server](https://github.com/pronovic/apologies-server) is a [Websocket](https://en.wikipedia.org/wiki/WebSocket) server interface used to interactively play a multi-player game using the [Apologies](https://github.com/pronovic/apologies) library.  The Apologies library implements a game similar to the [Sorry](https://en.wikipedia.org/wiki/Sorry!_(game)) board game.  \n\nIt was written as a learning exercise and technology demonstration effort, and serves as a complete example of how to manage a modern (circa 2020) Python project, including style checks, code formatting, integration with IntelliJ, [CI builds at GitHub](https://github.com/pronovic/apologies-server/actions), and integration with [PyPI](https://pypi.org/project/apologiesserver/) and [Read the Docs](https://apologies-server.readthedocs.io/en/stable/).  \n\nSee the [documentation](https://apologies-server.readthedocs.io/en/stable/design.html) for notes about the public interface and the event model.\n\nAs of this writing, the published PyPI project does not include a script to run\nthe server. The only way to run it is from the codebase, for local testing. See\nthe [developer](https://github.com/pronovic/apologies-server/blob/master/DEVELOPER.md#running-the-server) documentation\nat GitHub for more information.\n\nAs a technology demonstration effort, the Apologies Server is fairly\nsimplistic.  It runs as a single stateful process that maintains game state in\nmemory.  It cannot be horizontally scaled, and there is no option for an\nexternal data store.  There is also only limited support for authentication and\nauthorization - any player can register any handle that is not currently being\nused.  We do enforce resource limits (open connections, registered users,\nin-progress games) to limit the amount of damage abusive clients can do.\n',
    'author': 'Kenneth J. Pronovici',
    'author_email': 'pronovic@ieee.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/apologiesserver/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
